/** <p>
 *--------------------------------------------------------------------
 *
 * TDetachThread.h
 * 
 * Copyright (c) 2001 Kreatel Communications AB. All Rights Reserved
 * Author: Carl-Magnus Pettersson (carpe)
 *
 *--------------------------------------------------------------------
 */

#ifndef TASYNCH_THREAD_H
#define TASYNCH_THREAD_H

#include <pthread.h>
#include <assert.h>
#include "TVerboseException.h"

/**
 * This is a class that encapsulate a pthread. pthreads are more complicated
 * than one may think and therefore be sure to use this class properly.
 *
 * There are two ways of using this pthreads:
 * <ol>
 *   <li>DETACHED - Start the thread invoking TDetachThread::Start(), then let the thread
 *                  execute until it finishes or abort it from another thread
 *                  using TDetachThread::SAbort(). This will make the thread stop and the
 *                  memory is deallocated. </li>
 *   <li>JOINABLE - Start the thread invoking TJoinThread::Start(), then join the thread 
 *                  using TJoinThread::Join(). <b>NOTE</b> that join must be called, othervise the
 *                  memory will not be deallocated. </li>
 * </ol>              
 *
 */
class TDetachThread
{ 
private:  
  pthread_t Thread;
  pthread_attr_t Attributes;

  static void* Dispatch(void* arg);

protected:
  virtual void Execute() throw() = 0;

public:
  enum {
    THREAD_START_ERROR, 
    THREAD_ABORT_ERROR, 
    THREAD_JOIN_ERROR
  };

  TDetachThread();
  virtual ~TDetachThread();
  void Start() throw(TVerboseException);
  virtual void Abort() throw();
};

inline TDetachThread::TDetachThread()
{
  // Empty
}

inline TDetachThread::~TDetachThread()
{
  ::pthread_attr_destroy(&Attributes);
}

/**
 * Started as detached which means that it will
 * deallocate its memory. Should be called only 
 * from another thread.
 */ 
inline void TDetachThread::Start() throw(TVerboseException)
{
  ::pthread_attr_init(&Attributes);
  if (::pthread_attr_setdetachstate(&Attributes, 
                                    PTHREAD_CREATE_DETACHED) != 0) {
    throw TVerboseException("Error creating thread attributes", 
                            THREAD_START_ERROR);
  } 
  if (::pthread_create(&Thread, &Attributes, &Dispatch, this) != 0) {
    throw TVerboseException("Error starting thread", THREAD_START_ERROR);
  }
}

/**
 * Should be called only from another thread.
 */
inline void TDetachThread::Abort() throw()
{
  ::pthread_cancel(Thread);
}

inline void* TDetachThread::Dispatch(void* arg)
{
  ::pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
  reinterpret_cast<TDetachThread*>(arg)->Execute();
  return 0;
}
#endif 
